from .impl.dataset import DataSet
from .impl.decorators import with_project
from a2ml.api.utils.decorators import error_handler, authenticated
from .impl.cloud.rest_api import RestApi
from .credentials import Credentials
from .config import AugerConfig


class AugerDataset(object):
    def __init__(self, ctx):
        self.ctx = ctx
        self.credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            self.credentials.api_url, self.credentials.token)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def list(self, project):
        count = 0
        selected = self.ctx.config.get('dataset', None)
        for dataset in iter(DataSet(self.ctx, project).list()):
            self.ctx.log(
                ('[%s] ' % ('x' if selected == dataset.get('name') else ' ')) +
                dataset.get('name')
            )
            count += 1
        self.ctx.log('%s DataSet(s) listed' % str(count))
        return {'datasets': DataSet(self.ctx, project).list()}

    @error_handler
    @authenticated
    @with_project(autocreate=True)
    def create(self, project, source = None, validation=False, name=None, description=None):
        dataset = self._create(project, source, validation, name, description)
        self.ctx.log('Created DataSet %s' % dataset.name)
        return {'created': dataset.name}

    def _create(self, project, source = None, validation=False, name=None, description=None):
        if source is None:
            source = self.ctx.config.get('source', None)
        dataset = DataSet(self.ctx, project).create(source, name, description)
        AugerConfig(self.ctx).set_data_set(dataset.name, source, validation)

        return dataset

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def upload(self, project, source = None, name=None):
        data_source_file, local_data_source = \
            DataSet.verify(source, self.ctx.config.path)

        file_url, file_name = DataSet(self.ctx, project, name).do_upload_file(data_source_file, local_data_source=local_data_source)
        #self.ctx.log('Upload dataset %s to %s' % (name, file_url))
        return file_url

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def delete(self, project, name):
        if name is None:
            name = self.ctx.config.get('dataset', None)
        DataSet(self.ctx, project, name).delete()
        if name == self.ctx.config.get('dataset', None):
            AugerConfig(self.ctx).set_data_set(None, None, False).set_experiment(None, None)
        self.ctx.log('Deleted dataset %s' % name)
        return {'deleted': name}

    @error_handler
    def select(self, name):
        old_name = self.ctx.config.get('dataset', None)
        if name != old_name:
            AugerConfig(self.ctx).set_data_set(name, None, False).set_experiment(None, None)
        self.ctx.log('Selected DataSet %s' % name)
        return {'selected': name}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def download(self, project, name, path_to_download):
        if name is None:
            name = self.ctx.config.get('dataset', None)
        file_name = DataSet(self.ctx, project, name).download(path_to_download)
        self.ctx.log('Downloaded dataset %s to %s' % (name, file_name))
        return {'dowloaded': name, 'file': file_name}

    def preprocess_data(self, data, preprocessors, locally):
        if locally:
            return self._preprocess_data_locally(data, preprocessors)
        else:
           raise Exception("preprocess_data supported with locally=True only.")     

    def _preprocess_data_locally(self, data, preprocessors):                
        from auger_ml.preprocessors.text import TextPreprocessor

        res = data
        for p in preprocessors:
            name = list(p.keys())[0]
            params = list(p.values())[0]
            if name != 'text':
                raise Exception("Only text preprocessor supported.")

            tp = TextPreprocessor(params)
            res = tp.fit_transform(res)

        return res
            


class AugerConfig(object):
    def __init__(self, ctx):
        super(AugerConfig, self).__init__()
        self.ctx = ctx

    def set_data_set(self, name, source=None, validation=False):
        #TODO: add more providers later
        if validation:
            self.ctx.config.set('experiment/validation_dataset', name)
            if self.ctx.use_auger_cloud() and 'azure' in self.ctx.get_providers():
                self.ctx.config.set('experiment/validation_dataset', name, "azure")
        else:
            #print("set_data_set: %s"%self.ctx.use_auger_cloud())
            self.ctx.config.set('dataset', name)
            if self.ctx.use_auger_cloud() and 'azure' in self.ctx.get_providers():
                self.ctx.config.set('dataset', name, "azure")

        self.ctx.config.write_all()
        return self
            
    def set_experiment(self, experiment_name, experiment_session_id):
        self.ctx.config.set('experiment/name', experiment_name)
        self.ctx.config.set('experiment/experiment_session_id', experiment_session_id)
        self.ctx.config.write()

    def set_project(self, project_name):
        self.ctx.config.set('name', project_name)
        self.ctx.config.write()
        return self



from a2ml.api.auger.model import AugerModel
from a2ml.api.auger.dataset import AugerDataset
from a2ml.api.auger.experiment import AugerExperiment

class AugerA2ML(object):
    """Auger A2ML implementation."""

    def __init__(self, ctx):
        super(AugerA2ML, self).__init__()
        self.ctx = ctx

    def import_data(self, source=None, name=None, description=None):
        return AugerDataset(self.ctx).create(source=source, name=name, description=description)

    def preprocess_data(self, data, preprocessors, locally=False):
        return AugerDataset(self.ctx).preprocess_data(data, preprocessors, locally)

    def train(self):
        return AugerExperiment(self.ctx).start()

    def evaluate(self, run_id = None):
        return AugerExperiment(self.ctx).leaderboard(run_id)

    def deploy(self, model_id, locally=False, review=True, name=None, algorithm=None, score=None, data_path=None, metadata=None ):
        return AugerModel(self.ctx).deploy(model_id, locally, review, name, algorithm, score, data_path, metadata)

    def predict(self, model_id, filename, threshold=None, locally=False, data=None, columns=None, 
        predicted_at=None, output=None, no_features_in_result=None, score=False, score_true_data=None):
        return AugerModel(self.ctx).predict(
            model_id, filename, threshold, locally, data, columns, predicted_at, output, 
            no_features_in_result, score, score_true_data)

    def actuals(self, model_id, filename=None, data=None, columns=None, actuals_at=None, 
        actual_date_column=None, experiment_params=None, locally=False):
        return AugerModel(self.ctx).actuals(
            model_id, filename, data, columns, actuals_at, actual_date_column, experiment_params, locally)

    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        return AugerModel(self.ctx).delete_actuals(
            model_id, with_predictions, begin_date, end_date, locally)

    def review(self, model_id):
        return AugerModel(self.ctx).review(model_id)




class AzureA2ML(object):
    """Auger A2ML implementation."""

    def __init__(self, ctx):
        super(AzureA2ML, self).__init__()
        self.ctx = ctx

    def import_data(self, source=None):
        from a2ml.api.azure.dataset import AzureDataset

        return AzureDataset(self.ctx).create(source=source)

    def train(self):
        from a2ml.api.azure.experiment import AzureExperiment

        return AzureExperiment(self.ctx).start()

    def evaluate(self, run_id = None):
        from a2ml.api.azure.experiment import AzureExperiment

        return AzureExperiment(self.ctx).leaderboard(run_id)

    def deploy(self, model_id, locally=False, review=True, name=None, algorithm=None, score=None, data_path=None):
        from a2ml.api.azure.model import AzureModel

        return AzureModel(self.ctx).deploy(model_id, locally, review, name, algorithm, score, data_path)

    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, output=None, no_features_in_result=None):
        from a2ml.api.azure.model import AzureModel

        return AzureModel(self.ctx).predict(
            filename, model_id, threshold=threshold, locally=locally, data=data, columns=columns,
            predicted_at=predicted_at, output=output, no_features_in_result=no_features_in_result)

    def actuals(self, model_id, filename=None, data=None, columns=None, actuals_at=None, actual_date_column=None, locally=False):
        from a2ml.api.azure.model import AzureModel

        return AzureModel(self.ctx).actuals(
            model_id, filename, data, columns, actuals_at, actual_date_column, locally)

    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        from a2ml.api.azure.model import AzureModel

        return AzureModel(self.ctx).delete_actuals(
            model_id, with_predictions, begin_date, end_date, locally)

    def review(self, model_id):
        from a2ml.api.azure.model import AzureModel

        return AzureModel(self.ctx).review(model_id)




import os

class Config:
    def __init__(self):
        self.debug = os.environ.get('DEBUG', 'false').lower() == 'true'
        self.notificator_redis_host = os.environ.get('NOTIFICATOR_REDIS_HOST')
        self.notificator_redis_port = int(os.environ.get('NOTIFICATOR_REDIS_PORT', 6379))

        self.aws_access_key_id = os.environ.get('AWS_ACCESS_KEY_ID')
        self.aws_secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
        self.s3_endpoint_url = os.environ.get('S3_ENDPOINT_URL')
        self.upload_bucket = os.environ.get('UPLOAD_BUCKET', 'incoming')



from a2ml.api.base_a2ml import BaseA2ML
from a2ml.api.utils.show_result import show_result

class A2MLProject(BaseA2ML):
    """Contains the project CRUD operations that interact with provider."""
    def __init__(self, ctx, provider=None):
        """Initializes a new a2ml project.

        Args:
            ctx (object): An instance of the a2ml Context.
            provider (str): The automl provider(s) you wish to run. For example 'auger,azure,google'. The default is None - use provider set in config.
        
        Returns:
            A2MLProject object

        Examples:
            .. code-block:: python

                ctx = Context()
                project = A2MLDataset(ctx, 'auger, azure')
        """
        super(A2MLProject, self).__init__(ctx, 'project')
        self.runner = self.build_runner(ctx, provider)

    @show_result
    def list(self):
        """
        List all of the projects for the specified providers.

        Note:
            You will need to user the `iter <https://www.programiz.com/python-programming/methods/built-in/iter>`_ function to access the dataset elements.
        

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'projects': <object> 
                        }
                    }
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                project_list = A2MLProject(ctx, 'auger, azure').list()
                for provider in ['auger', 'azure']
                    if project_list[provider].result is True:
                        for project in iter(project_list[provider].data.projects):
                            ctx.log(project.get('name'))
                    else:
                        ctx.log('error %s' % project_list[provider].data)
        """
        return self.runner.execute('list')

    @show_result
    def create(self, name):
        """Creates a project for the specified providers.

        Args:
            name(str): name of project. If None - use project name from config.

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'created': 'project_name'
                    }
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                project_list = A2MLProject(ctx, 'auger, azure').create('new_project_name')


        """
        return self.runner.execute('create', name)

    @show_result
    def delete(self, name):
        """Deletes a project for the specified providers.

        Args:
            name(str): name of project. If None - use project name from config.

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'deleted': 'existing_project_name'
                    }
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                project_list = A2MLProject(ctx, 'auger, azure').delete('existinng_project_name')

        """
        return self.runner.execute('delete', name)

    @show_result
    def select(self, name):
        """Sets a Project name in the context.

        Args:
            name(str): name of project. name(str): name of project. If None - use project name from config.

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'selected': 'fortunetest'
                        }
                    }
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                DataSet(ctx, 'auger, azure').select(dataset_name)
        """
        return self.runner.execute('select', name)

    @show_result
    def get_cluster_config(self, name, local_config=True):
        """Get project cluster configuration for the specified providers.

        Args:
            name(str): name of project. If None - use project name from config.
            local_config(bool): If True, return cluster parameters from local config, otherwise from remote cluster

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'type': 'standard',
                            'min_nodes': 2,
                            'max_nodes': 2,
                            'stack_version': 'stable'
                        }
                    },
                    'azure': {
                        'result': True,
                        'data': {
                            'region': 'eastus2',
                            'min_nodes': 0,
                            'max_nodes': 2,
                            'type': 'STANDARD_D2_V2',
                            'name': 'a2ml-azure',
                            'idle_seconds_before_scaledown': 120
                        }
                    }                    
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                cluster_config = A2MLProject(ctx, 'auger, azure').get_cluster_config()

        """        
        return self.runner.execute('get_cluster_config', local_config)

    @show_result
    def update_cluster_config(self, name, params):
        """Update project cluster configuration for the specified providers.

        Args:
            name(str): name of project. If None - use project name from config.
            params(dict): cluster parameters to update.

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': None
                    },
                    'azure': {
                        'result': True,
                        'data': None
                    }                    
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                params = {'max_nodes': 4}
                cluster_config = A2MLProject(ctx, 'auger, azure').update_cluster_config(params)

        """                
        return self.runner.execute('update_cluster_config', name, params)



import os
import subprocess
import numpy as np

from ..cloud.cluster import AugerClusterApi
from ..cloud.pipeline import AugerPipelineApi
from ..cloud.experiment_session import AugerExperimentSessionApi
from ..cloud.endpoint import AugerEndpointApi
from ..cloud.review_alert import AugerReviewAlertApi
from ..cloud.review_alert_item import AugerReviewAlertItemApi
from ..exceptions import AugerException
from ..cloud.pipeline_file import AugerPipelineFileApi
from a2ml.api.utils import fsclient


class ModelDeploy(object):
    """Deploy Model on locally or on Auger Cloud."""

    def __init__(self, ctx, project):
        super(ModelDeploy, self).__init__()
        self.project = project
        self.ctx = ctx

    def execute(self, model_id, locally=False, review=True, name=None, algorithm=None, score=None, data_path=None, metadata=None):
        res = None
        if not locally or review:
            res = self.deploy_model_in_cloud(model_id, review, name, algorithm, score, data_path, metadata)

        if locally:
            res = self.deploy_model_locally(model_id, review, name, data_path, locally)

        return res

    def create_update_review_alert(self, model_id, pipeline_properties=None, parameters=None, name=None):
        if not self.ctx.config.get('review'):
            raise Exception("To use Review, please add section review to config.yml")

        if not pipeline_properties:
            pipeline_properties = AugerPipelineApi(self.ctx, None, model_id).properties()

        endpoint_api = None
        update_name = True
        if not pipeline_properties.get('endpoint_pipelines'):
            self.ctx.log('Creating review endpoint ...')
            endpoint_api = AugerEndpointApi(self.ctx, None)
            if not name:
                name = fsclient.get_path_base_name(self.ctx.config.get('source'))
            endpoint_properties = endpoint_api.create(pipeline_properties.get('id'), name)
            pipeline_properties['endpoint_pipelines'] = [endpoint_properties.get('id')]
            update_name = False

        if pipeline_properties.get('endpoint_pipelines'):
            if endpoint_api is None:
                endpoint_api = AugerEndpointApi(self.ctx, None, 
                    pipeline_properties['endpoint_pipelines'][0].get('endpoint_id'))

            params = {'review_metric': self.ctx.config.get('review/metric')}    
            if name and update_name:
                params['name'] = name

            if params:    
                endpoint_api.update(params)

            session_id = endpoint_api.properties().get('primary_experiment_session_id')
            if session_id:
                AugerExperimentSessionApi(self.ctx, None, None, session_id).update_settings()

            AugerReviewAlertApi(self.ctx, endpoint_api).create_update(parameters)

            endpoint_api.update_roi() 
            endpoint_api.update_monitoring_value()
        else:
            self.ctx.log('Model is not belong to any review endpoint. Skipping ...')

    def review(self, model_id):
        pipeline_properties = AugerPipelineApi(self.ctx, None, model_id).properties()
        result = {}
        if not pipeline_properties.get('endpoint_pipelines'):
            return result

        endpoint_api = AugerEndpointApi(self.ctx, None, 
            pipeline_properties['endpoint_pipelines'][0].get('endpoint_id'))

        alert_items = AugerReviewAlertItemApi(self.ctx, endpoint_api).list()
        if not alert_items:
            return result

        alert_items = sorted(alert_items, key=lambda k: k['id'], reverse=True)
        alert_item = alert_items[0]
        action_results = alert_item.get('action_results')

        error_states = ['review_data_build_failed', 'project_file_processing_failed', 'experiment_session_failed', 'pipeline_creating_failed']

        alert = AugerReviewAlertApi(self.ctx, None, alert_item['review_alert_id']).properties()
        status = "started"
        error = ""

        if alert.get('actions') == 'retrain_deploy':
            if action_results:
                retrain_status = action_results.get('retrain')
                if retrain_status in error_states:
                    status = 'error'
                    error = retrain_status
                else:
                    if retrain_status == 'external_pipeline_should_be_rebuilt':
                        status = 'retrain'
                    else:    
                        redeploy_status = action_results.get('redeploy')
                        if redeploy_status in error_states:
                            status = 'error'
                            error = redeploy_status
                        elif redeploy_status == 'endpoint_updated' or redeploy_status == 'endpoint_has_better_pipeline':
                            status = 'completed'
        elif alert.get('actions') == 'retrain':
            if action_results:
                retrain_status = action_results.get('retrain')
                if retrain_status in error_states:
                    status = 'error'
                    error = retrain_status
                elif retrain_status == 'experiment_session_done':
                    status = 'completed'
                elif retrain_status == 'external_pipeline_should_be_rebuilt':
                    status = 'retrain'
        else:
            status = 'completed'

        accuracy = 0
        if alert_item.get('values'):
            accuracy = np.mean(alert_item['values'])
        result = {
            'status': status,
            'error': error,
            'accuracy': accuracy
        }
        return result
            
    def deploy_model_in_cloud(self, model_id, review, name, algorithm, score, data_path, metadata):
        from .predict import ModelPredict

        self.ctx.log('Deploying model %s' % model_id)

        if self.ctx.is_external_provider():
            pipeline_properties = AugerPipelineApi(
                self.ctx, None).create_external(review, name, self.project.object_id, algorithm, score, metadata)
        else:    
            self.project.start()
            data_url = None
            if data_path:
                _, _, data_url, _ = ModelPredict(self.ctx)._process_input(data_path, None, None)

            pipeline_properties = AugerPipelineApi(
                self.ctx, None).create(model_id, review, name, data_url, metadata)

        if pipeline_properties.get('status') == 'ready':
            if review:
                self.create_update_review_alert(model_id, pipeline_properties, name=name)

            self.ctx.log('Deployed Model on Auger Cloud. Model id is %s' % \
                pipeline_properties.get('id'))            
        else:
            self.ctx.log('Deployed Model on Auger Cloud failed. Model id is %s.Error: %s' % \
                (pipeline_properties.get('id'), pipeline_properties.get('error_message', "")))

        return pipeline_properties.get('id')

    def deploy_model_locally(self, model_id, review, name, data_path, locally):
        is_loaded, model_path = self.verify_local_model(model_id)
        #TODO: support review flag
        if not is_loaded:
            self.ctx.log('Downloading model %s' % model_id)

            self.project.start()

            models_path = os.path.join(self.ctx.config.get_path(), 'models')
            pipeline_file_api = AugerPipelineFileApi(self.ctx, None)
            pipeline_file_properties = pipeline_file_api.create(model_id)
            downloaded_model_file = pipeline_file_api.download(
                pipeline_file_properties['signed_s3_model_path'],
                models_path, model_id)

            self.ctx.log('Downloaded model to %s' % downloaded_model_file)

            if locally == 'docker':
                self.ctx.log('Pulling docker image required to predict')
                self._docker_pull_image()
            else:
                self.ctx.log('To run predict locally install a2ml[predict]')             
        else:
            self.ctx.log('Downloaded model is %s' % model_path)

        return model_id

    def get_local_model_paths(self, model_id):
        models_path = os.path.join(self.ctx.config.get_path(), 'models')
        model_zip_path = os.path.join(models_path, 'model-%s.zip' % model_id)
        model_path = os.path.join(models_path,"model-%s"%model_id)

        return model_path, model_zip_path

    def verify_local_model(self, model_id, add_model_folder=True):
        model_path, model_zip_path = self.get_local_model_paths(model_id)

        is_exists = fsclient.is_folder_exists(model_path)
        if not is_exists and fsclient.is_file_exists(model_zip_path):
            self._extract_model(model_zip_path)

        if add_model_folder:    
            model_path = os.path.join(model_path, "model")

        is_exists = fsclient.is_folder_exists(model_path)

        return is_exists, model_path

    def _extract_model(self, model_name):
        from zipfile import ZipFile

        model_path = os.path.splitext(model_name)[0]
        model_existed = os.path.exists(model_path)

        if not model_existed:
            with ZipFile(model_name, 'r') as zip_file:
                zip_file.extractall(model_path)

        return model_path, model_existed

    def _docker_pull_image(self):
        cluster_settings = AugerClusterApi.get_cluster_settings(self.ctx)
        docker_tag = cluster_settings.get('kubernetes_stack')

        try:
            subprocess.check_call(
                'docker pull deeplearninc/auger-ml-worker:%s' % \
                 docker_tag, shell=True)
        except subprocess.CalledProcessError as e:
            raise AugerException('Can\'t pull Docker container...')

        return docker_tag



import asyncio
import contextlib
import json
import jsonpickle
import os
import requests
import sys
import time
import traceback
import websockets

from a2ml.api.utils import fsclient, dict_dig
from a2ml.api.a2ml_credentials import A2MLCredentials
from a2ml.api.utils.file_uploader import FileUploader, OnelineProgressPercentage
from a2ml.api.utils.provider_runner import ProviderRunner


class RequestMixin(object):
    def make_requst(self, http_verb, path, params={}):
        method = getattr(requests, http_verb)
        return self.handle_respone(method(f"{self.server_endpoint}{path}", json=params))

    def handle_respone(self, response):
        if response.status_code == 200:
            return response.json()
        else:
            self.show_output(f"Request error: {response.status_code} {response.text}")
            raise Exception(f"Request error: {response.status_code} {response.text}")

    def create_uploader(self):
        creds = self.get_upload_credentials()

        return FileUploader(
            bucket=creds['bucket'],
            endpoint_url=creds['endpoint_url'],
            access_key=creds['access_key'],
            secret_key=creds['secret_key'],
            session_token=creds['session_token'],
        )

    def get_upload_credentials(self):
        res = self.make_requst('post', '/api/v1/upload_credentials')
        return res['data']


class RemoteRunner(RequestMixin, ProviderRunner):
    def __init__(self, ctx, provider, obj_name = None):
        super().__init__(ctx, provider)

        self.obj_name = obj_name
        self.server_endpoint = ctx.config.get('server_endpoint', os.environ.get('A2ML_SERVER_ENDPOINT'))
        self.ws_endpoint = 'ws' + self.server_endpoint[4:]

    def execute(self, operation_name, *args, **kwargs):
        new_args, uploaded_file, local_file = self.upload_local_files(operation_name, *args, **kwargs)
        self.uploaded_file = uploaded_file
        self.local_file = local_file

        return super().execute(operation_name, *new_args['args'], **new_args['kwargs'])

    def execute_provider(self, provider_name, operation_name, *args, **kwargs):
        ctx = self.ctx.copy(provider_name)
        provider_runner = RemoteProviderRunner(ctx, provider_name, self.obj_name)
        return provider_runner.execute(operation_name, self.uploaded_file, self.local_file, *args, **kwargs)

    def upload_local_files(self, operation_name, *args, **kwargs):
        file_to_upload = None
        inject_uploaded_file_to_request_func = None

        if operation_name == 'import_data':
            file_to_upload = kwargs.get('source') or self.ctx.config.get('source')

            def replacer_func(remote_path):
                new_kwargs = dict(kwargs)
                new_kwargs['source'] = remote_path
                return {'args': args, 'kwargs': new_kwargs}

            inject_uploaded_file_to_request_func = replacer_func
        elif operation_name == 'predict' or (self.obj_name == 'dataset' and operation_name == 'create'):
            file_to_upload = args[0]

            def replacer_func(remote_path):
                new_args = list(args)
                new_args[0] = remote_path
                return {'args': tuple(new_args), 'kwargs': kwargs}

            inject_uploaded_file_to_request_func = replacer_func

        if file_to_upload and self.local_path(file_to_upload):
            uploader = self.create_uploader()
            url = uploader.multi_part_upload(file_to_upload, callback=OnelineProgressPercentage(file_to_upload))

            return inject_uploaded_file_to_request_func(url), url, file_to_upload
        else:
            return {'args': args, 'kwargs': kwargs}, None, None

    def local_path(self, path):
        path = path.lower()
        return not(path.startswith('s3://') or path.startswith('http://') or path.startswith('https://'))

class RemoteProviderRunner(RequestMixin, object):
    CRUD_TO_METHOD = {
        'create': 'post',
        'delete': 'delete',
        'list': 'get',
    }

    NON_CRUD_TO_METHOD = {
        'actuals': 'post',
        'deploy': 'patch',
        'evaluate': 'patch',
        'history': 'get',
        'import_data': 'patch',
        'leaderboard': 'get',
        'predict': 'post',
        'review': 'patch',
        'select': 'patch',
        'start': 'patch',
        'stop': 'patch',
        'train': 'patch',
    }

    def __init__(self, ctx, provider, obj_name):
        super().__init__()

        self.ctx = ctx
        self.provider = provider
        self.obj_name = obj_name
        self.server_endpoint = ctx.config.get('server_endpoint', os.environ.get('A2ML_SERVER_ENDPOINT'))
        self.ws_endpoint = 'ws' + self.server_endpoint[4:]

    def execute(self, operation_name, uploaded_file, local_file, *args, **kwargs):
        try:
            creds = A2MLCredentials(self.ctx, self.provider)
            creds.load()
            creds.verify()
            self.ctx.credentials = creds.serialize()

            http_verb, path = self.get_http_verb_and_path(operation_name)
            # new_args, uploaded_file, local_file = self.upload_local_files(operation_name, *args, **kwargs)
            # params = self._params(*new_args['args'], **new_args['kwargs'])

            params = self._params(*args, **kwargs)

            if uploaded_file:
                params['tmp_file_to_remove'] = uploaded_file

            res = self.make_requst(http_verb, path, params)

            with self.get_event_loop() as event_loop:
                return event_loop.run_until_complete(self.wait_result(res['data']['request_id'], local_file))
        except Exception as exc:
            if self.ctx.debug:
                traceback.print_exc()

            self.ctx.log(str(exc))

            return {
                self.provider: {
                    'result': False,
                    'data': str(exc)
                }
            }

    def _params(self, *args, **kwargs):
        return {
            'context': jsonpickle.encode(self.ctx),
            'project_name': self.ctx.config.get('name'),
            'args': args,
            'kwargs': kwargs,
        }

    def get_http_verb_and_path(self, operation_name):
        crud = False

        if operation_name in self.CRUD_TO_METHOD:
            http_verb = self.CRUD_TO_METHOD[operation_name]
            crud = True
        elif operation_name in self.NON_CRUD_TO_METHOD:
            http_verb = self.NON_CRUD_TO_METHOD[operation_name]
        else:
            raise ValueError(f'unknown operation {operation_name}')

        if self.obj_name:
            if crud:
                path = f'/api/v1/{self.obj_name}s'
            else:
                path = f'/api/v1/{self.obj_name}s/{operation_name}'
        else:
            path = f'/api/v1/{operation_name}'

        return (http_verb, path)

    @contextlib.contextmanager
    def get_event_loop(self):
        try:
            yield asyncio.get_event_loop()
        except RuntimeError:
            try:
                loop = asyncio.new_event_loop()
                yield loop
            finally:
                loop.close()

    def handle_weboscket_respone(self, data, local_file):
        data_type = self.get_response_data_type(data)

        if data_type == 'result' and isinstance(data['result'], dict):
            config_json = data['result'].get('config')
            if config_json:
                config = jsonpickle.decode(config_json)
                config.write_all()

            response = data['result'].get('response')
            if response:
                self.download_prediction_result(response, local_file)

        self.show_output(data)

    def download_prediction_result(self, results, local_file):
        for provider in results.keys():
            result = results[provider]
            predicted = dict_dig(result, 'data', 'predicted')
            if predicted and type(predicted)==str and fsclient.is_s3_path(predicted):
                file_path = os.path.splitext(local_file)[0] + '_predicted.csv'
                client = self.create_uploader()
                client.download(predicted, file_path)
                result['data']['predicted'] = file_path
                self.ctx.log(f'{predicted} downloaded to {file_path}')

    def get_response_data_type(self, data):
        if isinstance(data, dict):
            return data.get('type', None)

    def show_output(self, data):
        data_type = self.get_response_data_type(data)

        if data_type == 'log':
            self.ctx.log(str(data['msg']))
        elif data_type == 'result':
            if self.ctx.debug or data.get('status') == 'FAILURE':
                self.ctx.log(str(data['status']) + ': ' + str(dict_dig(data, 'result', 'response')))
        elif data_type == 'ping':
            pass
        elif data_type == 'start':
            pass
        elif isinstance(data, Exception):
            if self.ctx.debug:
                traceback.print_exc()

            self.ctx.log('Error: ' + str(data))
        else:
            self.ctx.log(str(data))

    async def wait_result(self, request_id, local_file):
        data = {}

        while not self.get_response_data_type(data) == 'result':
            data = await self.connect_and_get_result(request_id, local_file)

        if data.get('status') == 'SUCCESS':
            result = data.get('result', {})
            return dict_dig(result, 'response') or result
        else:
            return {
                self.provider: {
                    'result': False,
                    'data': "Server error. Details: %s"%data.get('result')
                }
            }

    async def connect_and_get_result(self, request_id, local_file):
        last_msg_id = '0'
        try:
            endpoint = self.ws_endpoint + "/ws?id=" + request_id + '&last_msg_id=' + str(last_msg_id)
            async with websockets.connect(endpoint) as websocket:
                while True:
                    data = await websocket.recv()
                    data = json.loads(data)

                    if '_msg_id' in data:
                        last_msg_id = data['_msg_id']

                    self.handle_weboscket_respone(data, local_file)

                    if self.get_response_data_type(data) == 'result':
                        return data
        except Exception as e:
            self.show_output(e)
            await asyncio.sleep(2)
            return {}



import os

from .mparts.deploy import ModelDeploy
from .mparts.undeploy import ModelUndeploy
from .mparts.predict import ModelPredict
from .mparts.actual import ModelActual
from .mparts.delete_actual import ModelDeleteActual
from a2ml.api.model_review.model_review import ModelReview
from .exceptions import AugerException
from .cloud.endpoint import AugerEndpointApi
from .cloud.pipeline import AugerPipelineApi

class Model(object):
    """Auger Cloud Model(s) management."""

    def __init__(self, ctx, project):
        super(Model, self).__init__()
        self.project = project
        self.ctx = ctx

    def deploy(self, model_id, locally=False, review=True, name=None, algorithm=None, score=None, data_path=None, metadata=None):
        return ModelDeploy(self.ctx, self.project).execute(model_id, locally, review, name, algorithm, score, data_path, metadata)

    def review_alert(self, model_id, parameters, name):
        return ModelDeploy(self.ctx, self.project).create_update_review_alert(model_id, None, parameters, name)

    def review(self, model_id):
        return ModelDeploy(self.ctx, self.project).review(model_id)

    def undeploy(self, model_id, locally=False):
        return ModelUndeploy(self.ctx, self.project).execute(model_id, locally)

    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, 
        output=None, no_features_in_result=None, score=False, score_true_data=None):
        return ModelPredict(self.ctx).execute(filename, model_id, threshold, locally, data, columns, 
            predicted_at, output, no_features_in_result, score, score_true_data)

    def actuals(self, model_id, filename=None, data=None, columns=None, actuals_at=None, 
        actual_date_column=None, experiment_params=None, locally=False):
        if locally:
            is_loaded, model_path = ModelDeploy(self.ctx, self.project).verify_local_model(model_id)

            if not is_loaded:
                raise AugerException('Model should be deployed locally.')

            params = {
                'model_path': model_path,
                'roi': {
                    'filter': str(self.ctx.config.get('review/roi/filter')),
                    'revenue': str(self.ctx.config.get('review/roi/revenue')),
                    'investment': str(self.ctx.config.get('review/roi/investment')),
                }
            }
            return ModelReview(params).add_actuals(
              self.ctx,
              actuals_path=filename,
              data=data,
              columns=columns,
              actual_date=actuals_at,
              actual_date_column=actual_date_column,
              experiment_params=experiment_params,
              provider='auger'
            )
        else:
            return ModelActual(self.ctx).execute(model_id, filename, data, columns, actuals_at, 
                actual_date_column, experiment_params)

    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        if locally:
            is_loaded, model_path = ModelDeploy(self.ctx, self.project).verify_local_model(model_id)
            if not is_loaded:
                raise AugerException('Model should be deployed locally.')

            return ModelReview({'model_path': model_path}).delete_actuals(
              with_predictions=with_predictions, begin_date=begin_date, end_date=end_date)
        else:
            return ModelDeleteActual(self.ctx).execute(model_id, with_predictions, begin_date, end_date)

    def build_review_data(self, model_id, locally, output):
        if locally:
            is_loaded, model_path = ModelDeploy(self.ctx, self.project).verify_local_model(model_id)
            if not is_loaded:
                raise AugerException('Model should be deployed locally.')

            return ModelReview({'model_path': model_path}).build_review_data(
              data_path=self.ctx.config.get("source"), output=output)
        else:
            raise Exception("Not Implemented.")

    def get_info(self, model_id, locally):
        if locally:
            raise AugerException('Model get_info for local model is not supported yet.')

        pipeline_properties = AugerPipelineApi(self.ctx, None, model_id).properties()    

        return pipeline_properties

    def update(self, model_id, updates, locally):
        if locally:
            raise AugerException('Model get_info for local model is not supported yet.')

        return AugerPipelineApi(self.ctx, None, model_id).update(updates)    



import os
from pathlib import Path
from a2ml.cmdl.cmdl import cmdl


class TestCmdl(object):

    def test_empty_a2ml(self, runner):
        result = runner.invoke(cmdl)
        assert result.exit_code == 0
        assert 'A2ML command line interface.' in result.output
        assert 'auth        Authenticate with AutoML provider.' in result.output
        assert 'deploy      Deploy trained model.' in result.output
        assert 'evaluate    Evaluate models after training.' in result.output
        assert 'import      Import data for training.' in result.output
        assert 'new         Create new A2ML project.' in result.output
        assert 'predict     Predict with deployed model.' in result.output
        assert 'train       Train the model.' in result.output


class TestNewCmd(object):
    PROJECT_NAME = 'new_project'

    def test_new_requires_project_name(self, runner):
        result = runner.invoke(cmdl, ['new'])
        assert 'Error: Missing argument' in result.output
        assert 'PROJECT' in result.output
        assert result.exit_code == 2

    def test_new_creates_valid_structure(self, runner):
        with runner.isolated_filesystem():
            result = runner.invoke(cmdl, ['new', self.PROJECT_NAME])
            assert result.exit_code == 0
            assert Path(self.PROJECT_NAME).exists()
            for filename in ('google.yaml', 'azure.yaml', 'config.yaml', 'auger.yaml'):
                file = Path(os.path.join(self.PROJECT_NAME, filename))
                assert file.exists()

    def test_new_already_existing(self, runner, log):
        with runner.isolated_filesystem():
            result = runner.invoke(cmdl, ['new', self.PROJECT_NAME])
            assert result.exit_code == 0
            result = runner.invoke(cmdl, ['new', self.PROJECT_NAME])
            assert result.exit_code == 0
            assert (log.messages[-1] ==
                "[config] Can't create 'new_project'. Folder already exists.")
